from asyncio import AbstractEventLoop, Future, get_event_loop
from logging import Logger
from typing import Callable, Coroutine, Dict, List, Optional

from ..base import get_logger

__all__ = ("Listener",)

log: Logger = get_logger("dispatch")


class Listener:
    """
    A class representing how events become dispatched and listened to.

    :ivar AbstractEventLoop loop: The coroutine event loop established on.
    :ivar dict events: A list of events being dispatched.
    """

    __slots__ = ("loop", "events", "extra_events")

    def __init__(self) -> None:
        self.loop: AbstractEventLoop = get_event_loop()
        self.events: Dict[str, List[Callable[..., Coroutine]]] = {}
        self.extra_events: Dict[str, List[Future]] = {}  # used in `Client.wait_for`

    def dispatch(self, name: str, /, *args, **kwargs) -> None:
        r"""
        Dispatches an event given out by the gateway.

        :param str name: The name of the event to dispatch.
        :param list[Any] \*args: Multiple arguments of the coroutine.
        :param dict \**kwargs: Keyword-only arguments of the coroutine.
        """
        for event in self.events.get(name, []):
            converters: dict
            if converters := getattr(event, "_converters", None):
                _kwargs = kwargs.copy()
                for key, value in _kwargs.items():
                    if key in converters.keys():
                        del kwargs[key]
                        kwargs[converters[key]] = value

            self.loop.create_task(event(*args, **kwargs))
            log.debug(f"DISPATCH: {event}")

        # wait_for events
        futs = self.extra_events.get(name, [])
        if not futs:
            return

        log.debug(f"Resolving {len(futs)} futures")

        for fut in futs:
            if fut.done():
                log.debug(
                    f"A future for the {name} event was already {'cancelled' if fut.cancelled() else 'resolved'}"
                )
            else:
                fut.set_result(args)

        self.extra_events[name] = []

    def register(self, coro: Callable[..., Coroutine], name: Optional[str] = None) -> None:
        """
        Registers a given coroutine as an event to be listened to.
        If the name of the event is not given, it will then be
        determined by the coroutine's name.

        i.e. : async def on_guild_create -> "ON_GUILD_CREATE" dispatch.

        :param Callable[..., Coroutine] coro: The coroutine to register as an event.
        :param Optional[str] name: The name to associate the coroutine with. Defaults to None.
        """
        _name: str = coro.__name__ if name is None else name
        event = self.events.get(_name, [])
        event.append(coro)

        self.events[_name] = event
        log.debug(f"REGISTER: {self.events[_name]}")

    def add(self, name: str) -> Future:
        """
        Returns a Future that will resolve whenever the supplied event is dispatched

        :param str name: The event to listen for
        :return: A future that will be resolved on the next event dispatch with the data given
        :rtype: asyncio.Future
        """
        fut = self.loop.create_future()
        futures = self.extra_events.get(name, [])
        futures.append(fut)
        self.extra_events[name] = futures
        return fut
